import numpy as np
import sympy
import scipy.sparse

import kwant.kpm
from kwant._common import ensure_rng

from qsymm.model import Model, _symbol_normalizer

from .higher_order_lowdin import _interesting_keys, effective_model
from .kpm_funcs import _kpm_preprocess
from .hybrid_kpm_jj import sparse_diag

one = sympy.sympify(1)


def trace_perturbation(H0, H1, order=2, interesting_keys=None,
                       trace=True, num_moments=100, operator=None,
                       num_vectors=10, vectors=None,
                       normalized_vectors=False, kpm_params=dict()
                      ):
    """
    Perturbative expansion of `Tr(f(H) * A)` using stochastic trace.

    Parameters
    ----------

    H0 : array or Model
        Unperturbed hamiltonian, dense or sparse matrix. If
        provided as Model, it must contain a single
        entry {1 : array}.
    H1 : dict of {symbol : array} or Model
        Perturbation to the Hamiltonian.
    order : int (default 2)
        Order of the perturbation calculation, up to `order` order
        polynomials of every key in `H1` is kept during the expansion.
        If `interesting_keys` is provided, `order` is ignored.
    interesting_keys : iterable of sympy expressions or None (default)
        The calculation can be sped up by providing an explicit set of keys
        to keep. Should contain all subexpressions of desired keys, as terms
        not listed in `interesting_keys` are discarded at every step of the
        calculation.
    trace : bool, default True
        Whether to calculate the trace. If False, all matrix elements
        between pairs of vectors are separately returned.
    num_moments : int, default 100
        Number of moments in the KPM expansion.
    operator : 2D array or Model or None (default)
        Operator in the expectation value, default is identity.
    num_vectors : int, default 10
        Number of random vectors used in KPM. Ignored if `vector_factory`
        is provided.
    vectors : 1D or 2D array or Model or None (default)
        Vector of length `N` or array of vectors with shape `(M, N)`.
        If Model, must be of shape `(M, N)`. The size of
        the last index should be the same as the size of the Hamiltonian.
        By default, random vectors are used.
    normalized_vectors: bool, default False
        Whether the vectors returned by `vector_factory` are normalized.
        By default the result is divided by the number of vectors, as
        random phase vectors are assumed. If True, the result is not
        divided.
    kpm_params : dict, optional
        Dictionary containing additional KPM parameters, see `~kwant.kpm`.

    Returns
    -------

    expansion : callable
        A function that takes the function of energy `f` and returns the
        expansion of the trace. This evaluation is very fast, as the
        moments of the expansion are precalculated.
    """
    H1 = Model(H1)
    all_keys = _interesting_keys(H1.keys(), order)
    if interesting_keys is None:
        interesting_keys = all_keys
    else:
        interesting_keys = {_symbol_normalizer(k) for k in interesting_keys}
    H1.keep = interesting_keys
    # Convert to appropriate format
    if not isinstance(H0, Model):
        H0 = Model({1: H0}, keep=interesting_keys)
    elif not (len(H0) == 1 and list(H0.keys()).pop() == 1):
        raise ValueError('H0 must contain a single entry {sympy.sympify(1): array}.')
    # Find the bounds of the spectrum and rescale `ham`
    kpm_params['num_moments'] = num_moments
    H0[one], (_a, _b), num_moments, _kernel = _kpm_preprocess(H0[one], kpm_params)
    H1 /= _a

    ham = H0 + H1
    N = ham.shape[0]

    if vectors is None:
        # Make random vectors
        rng = ensure_rng(kpm_params.get('rng'))
        vectors = np.exp(2j * np.pi * rng.random_sample((num_vectors, N)))
    if not isinstance(vectors, Model):
        vectors = Model({1: np.atleast_2d(vectors)})

    # Calculate all the moments, this is where most of the work is done.
    moments = []
    # Precalculate operator acting on vectors, assume it is Hermitian
    if operator is not None:
        op_vecs = (operator @ vectors.T()).T()
    else:
        op_vecs = vectors

    for kpm_vec in _perturbative_kpm_vectors(ham, vectors, num_moments):
        next_moment = op_vecs.conj() @ kpm_vec.T()
        if trace:
            next_moment = next_moment.trace()
        if not normalized_vectors:
            next_moment /= num_vectors
        moments.append(next_moment)

    def expansion(f):
        """
        Perturbative expansion of `Tr(f(H) * A)` using stochastic trace.
        The moments are precalculated, so evaluation should be fast.
        """
        coef = np.polynomial.chebyshev.chebinterpolate(lambda x: f(x * _a + _b), num_moments)
        coef = _kernel(coef)
        return sum(c * moment for c, moment in zip(coef, moments))

    return expansion


def _perturbative_kpm_vectors(ham, vectors, max_moments):
    """
    Generator object that succesively yields KPM vectors `T_n(ham) |vector>`
    for vectors in `vectors` for n in [0, max_moments].

    Parameters
    ----------
    ham : Model
        Hamiltonian, Model with shape `(N, N)`.
        The `keep` property is used to limit the powers of
        free parameters that we keep track of in the expansion.
    vectors : 1D or 2D array or Model
        Vector of length `N` or array of vectors with shape `(M, N)`.
        If Model, must be of shape `(M, N)`. The size of
        the last index should be the same as the size of `ham` `N`.
    max_moments : int
        Number of moments to stop with iteration

    Notes
    -----
    Returns a sequence of expanded vectors, Models of shape (M, N).
    If the input was a vector, M=1.
    """
    if not isinstance(vectors, Model):
        alpha = Model({1: np.atleast_2d(vectors)})
    else:
        alpha = vectors
    # Internally store as column vectors
    alpha = alpha.T()
    n = 0
    yield alpha.T()
    n += 1
    alpha_prev = alpha
    alpha = ham @ alpha
    yield alpha.T()
    n += 1
    while n < max_moments:
        alpha_next = 2 * ham @ alpha - alpha_prev
        alpha_prev = alpha
        alpha = alpha_next
        yield alpha.T()
        n += 1


def hybrid_expansion(H0, Hpert,
                     df, f_expansion,
                     vectors,
                     A_window, B_window, sigma, k=20,
                     interesting_keys=None,
                     num_moments=100,
                     num_moments_Lowdin=100,
                     bounds=None,
                     separate_terms=False):
    """Second order hybrid perturbation expansion of Tr(f(H0 + Hpert)).
    It only contains linear and quadratic terms in the small parameters,
    not the constant term.

    Parameters:
    -----------
    H0 : array or Model
        Unperturbed hamiltonian, dense or sparse matrix. If
        provided as Model, it must contain a single
        entry {1 : array}.
    Hpert : dict of {symbol : array} or Model
        Perturbation to the Hamiltonian.
    df : callable
        Derivative of the function f to evaluate. Must be able to take a vector
        of values as argument.
    f_expansion : callable
        `f_expansion(x0, dx)` is the Taylor expansion of f around scalar
        x0 for small dx. Must be able to take arrays and Models as arguments
        to evaluate matrix polynomials.
    vectors : 2D array
        Vectors used to calculate the trace by summing over diagonal matrix elements.
        Array of vectors with shape `(M, N)`. The size of the last index should
        be the same as the size of the Hamiltonian.
    A_window, B_window, sigma : floats
        Used to find the A and B exact eigenvectors of `ham`. The A group is within
        `A_window` to `sigma` in energy, the B group is within `B_window`, but not
        in the first group.
    k : int
        At most `k` exact eigenvectors are found, if `k=0` all eigenvectors
        are found using full diagonalization.
    interesting_keys : iterable of sympy expressions or None (default)
        The calculation can be sped up by providing an explicit set of keys
        to keep. Should contain all subexpressions of desired keys, as terms
        not listed in `interesting_keys` are discarded at every step of the
        calculation.
    num_moments : int, default 100
        Number of moments in the KPM expansion.
    num_moments_Lowdin : int, default 100
        Number of moments in the Lowdin expansion.
    bounds : tuple of two floats or None (default)
        Bounds of the spectrum of `H0`.
    separate_terms : bool (default False)
        Whether to return the combined result or split the three contributions.
    """
    Hpert = Model(Hpert)
    all_keys = _interesting_keys(Hpert.keys(), 2)
    if interesting_keys is None:
        interesting_keys = all_keys
    else:
        interesting_keys = {_symbol_normalizer(k) for k in interesting_keys}
    Hpert.keep = interesting_keys

    # full KPM only
    exp_kpm_full = kpm_expansion(H0, Hpert, df, num_moments, vectors,
                                 bounds=bounds,
                                 interesting_keys=interesting_keys,)
    # subgap states
    evecA, evecB = exact_vectors(H0,
                             A_window=A_window,
                             B_window=B_window,
                             sigma=sigma,
                             k=k)
    # Lowdin effective Hamiltonain
    effH = combined_effective_model(H0, Hpert, evecA, evecB,
                                    interesting_keys=interesting_keys,
                                    num_moments_Lowdin=num_moments_Lowdin)
    # KPM contribution of subgap states
    exp_kpm_sg = kpm_expansion(effH[1], effH - effH[1], df, num_moments, vectors=np.eye(effH.shape[0]),
                               interesting_keys=interesting_keys, bounds=bounds)
    # Exact contribution of subgap states
    exp_exact_sg = f_expansion(effH[1], effH - effH[1]).trace()
    # Remove constant term
    exp_exact_sg -= exp_exact_sg[1]
    if separate_terms:
        return exp_kpm_full, exp_kpm_sg, exp_exact_sg
    else:
        return exp_kpm_full - exp_kpm_sg + exp_exact_sg


def kpm_expansion(H0, Hpert, df, num_moments, vectors, interesting_keys, bounds=None):
    """Generate first and second order KPM expansion of 'H0 + Hpert' using permutation of
    terms inside the trace, such that `vectors` can be chosen as a basis of the image space
    of `Hpert`."""

    expansion1 = trace_perturbation(H0,
        Model({1: Model({1: H0}).__missing__('x')}),
        operator=Model({'phi**2': Hpert['phi**2']}),
        order=0,
        interesting_keys=interesting_keys,
        num_moments=num_moments,
        vectors=vectors,
        normalized_vectors=True,
        kpm_params=dict(bounds=bounds)
    )
    expansion2 = trace_perturbation(
        H0, Model({'phi': Hpert['phi']}),
        operator=Model({'phi': Hpert['phi']}),
        order=1,
        interesting_keys=interesting_keys,
        num_moments=num_moments,
        vectors=vectors,
        normalized_vectors=True,
        kpm_params=dict(bounds=bounds)
    )
    exp1 = expansion1(df)
    exp2 = expansion2(df)
    first_order = exp2['phi']
    second_order = 2 * exp1['phi**2'] + exp2['phi**2']
    return Model({'phi': first_order, 'phi**2': (1/2) * second_order})


def exact_vectors(ham, A_window, B_window, k=20, sigma=0):
    """Return two groups of exact eigenvectors of `ham`, the first group is within
    `A_window` to `sigma` in energy, the second group is within `B_window`, but not
    in the first group. At most `k` eigenvectors are found, if `k=0` all eigenvectors
    are found."""
    if k == 0:
        # Full diagonalization
        ham = np.array(ham)
        evals, evecs = np.linalg.eigh(ham)
    else:
        evals, evecs = sparse_diag(ham, k=k, sigma=sigma, return_eigenvectors=True)
    evecs, _ = scipy.linalg.qr(evecs, mode ='economic')
    assert np.allclose(evecs.T.conj() @ evecs, np.eye(evecs.shape[-1]))
    maskA = abs(evals - sigma) <= A_window
    maskB = abs(evals - sigma) <= B_window
    maskB = np.logical_xor(maskA, maskB)
    return evecs[:, maskA], evecs[:, maskB]


def block_diag_model(*models):
    """Combine models to a block-diagonal Model"""
    result = Model({
    key: scipy.linalg.block_diag(*[model[key] for model in models])
    for key in set.union(*[set(model.keys()) for model in models])
    })
    return result


def combined_effective_model(H0, Hpert, evecA, evecB, interesting_keys, num_moments_Lowdin=1000):
    """Do Lowdin for every eigenstate in evecA separately, return the result as a
    diagonal model. Only works with nondegenerate spectrum at the moment.
    """
    H1 = Model(Hpert)
    terms = []
    for i in range(evecA.shape[-1]):
        vecA = evecA[:, i]
        vecsB = np.hstack([evecB, evecA[:, np.arange(evecA.shape[-1]) != i]])
        term = effective_model(H0, H1, vecA[:, None], vecsB, order=2,
                               interesting_keys=interesting_keys,
                               kpm_params=dict(num_moments=num_moments_Lowdin,
                               # kernel=kwant.kpm.jackson_kernel
                                              )
        )
        terms.append(term.toarray())
    return block_diag_model(*terms)


def kpm_A_expansion(H0, Hpert, f, evecA, evecB, num_moments=100, num_moments_Lowdin=1000, bounds=None):
    """Use Lowdin perturbation theory on evecA states
    to estimate contribution of these states to Tr(f(H)).
    Only works with nondegenerate spectrum at the moment."""
    H1 = Model(Hpert)
    effH = combined_effective_model(H0, Hpert, evecA, evecB, num_moments_Lowdin)
    exp = kpm_expansion(effH[1], effH - effH[1], f, num_moments, vectors=np.eye(effH.shape[0]), bounds=bounds)
    return exp
