import kwant
import numpy as np
import math
import scipy
import scipy.sparse.linalg as sla
from scipy.sparse import identity
import kwant.linalg.mumps as mumps
from numpy.polynomial import Chebyshev as T1

s0 = np.array([[1, 0], [0, 1]])
sx = np.array([[0, 1], [1, 0]])
sy = np.array([[0, -1j], [1j, 0]])
sz = np.array([[1, 0], [0, -1]])

def square_jj(a, LN, LS, W, finalized=True):
    """Make a Josephson junction of normal-region length LN,
    superconducting contact length LS, and width W.

    Parameters
    ----------
    LN : integer
        Length of the normal scattering region.
    LS : integer
        Length of the superconducting leads.
    W  : integer
        Width of the junction
    """

    def shape(pos):
        (x, y) = pos
        return -LS <= x <= LN + LS and 0 <= y <= W

    def onsite(site, t, mu, delta, phi):
        H0 = (4 * t - mu) * sz
        phase = delta * np.exp(1j * phi)
        ss = np.array([[0, phase], [phase.conj(), 0]])
        (x, y) = site.pos
        if 0 <= x <= LN:
            return H0
        elif x > LN:
            return H0 + ss
        return H0 + delta * sx

    def hopx(site1, site2, t):
        return -t * sz

    lat = kwant.lattice.square(norbs=2)
    sys = kwant.Builder()
    sys[lat.shape(shape, (0, 0))] = onsite
    sys[lat.neighbors()] = hopx
    if finalized:
        sys = sys.finalized()
    return sys

def sparse_diag(matrix, k, sigma, **kwargs):
    """Call sla.eigsh with mumps support.

    Please see scipy.sparse.linalg.eigsh for documentation.
    """

    class LuInv(sla.LinearOperator):
        def __init__(self, A):
            inst = mumps.MUMPSContext()
            inst.analyze(A, ordering="pord")
            inst.factor(A)
            self.solve = inst.solve
            sla.LinearOperator.__init__(self, A.dtype, A.shape)

        def _matvec(self, x):
            return self.solve(x.astype(self.dtype))

    opinv = LuInv(matrix - sigma * identity(matrix.shape[0]))
    return sla.eigsh(matrix, k, sigma=sigma, OPinv=opinv, **kwargs)

def andreev_vectors(syst, params, k=4, sigma=0.0):
    """Return eigenvalues and eigenvectors inside the  superconducting gap.
    """
    ham = syst.hamiltonian_submatrix(params=params, sparse=True)
    evals, evecs = sparse_diag(ham, k=k, sigma=sigma, return_eigenvectors=True)
    mask = abs(evals - sigma) <= params['delta']
    return evals[mask], evecs[:, mask]

def supercurrent_exact(syst, params, eigenvecs=None, cut=None):
    """Exact calculation of the supercurrent for every eigenvector
    or calculate total supercurrent when eigenvectors are not provided.
    """
    kwant_operator = kwant.operator.Current(syst, sz, where=cut, sum=True)
    kwant_operator = kwant_operator.bind(params=params)
    if eigenvecs is None:
        ham = syst.hamiltonian_submatrix(params=params)
        eigvals, eigenvecs = scipy.linalg.eigh(ham)
        eigenvecs = eigenvecs[:, eigvals < 0]
        # Make sure degeneracies don't cause problems
        eigenvecs, _ = scipy.linalg.qr(eigenvecs, mode ='economic')
    current = np.sum([kwant_operator(eigvec) for eigvec in eigenvecs.T])
    return current

def nth_moment(n, eigvals, current_op_expectation):
    """Returns nth coefficient of Chebyshev expansion of the current operator
    evaluated at given eigenvectors and eigenvalues.

    n: number (nth coefficient of Chebyshev expansion)
    eigvals: eigenvalues at which the Chebyshev expansion is made
    current_op_expectation: expectation value of current operator <k|I|k>
    with eigenvector |k> corresponding to kth eigen value.
    """
    tn = np.array([T1.basis(n)(eigval) for eigval in eigvals])
    return current_op_expectation @ tn

def current_spectrum(energies, current_operator, eigen_values, eigen_vectors, num_moments, bounds):
    """Returns the specteral density of current operator as a
    function of enrgies using Chebyshev expansion of current
    operator at given eigenvalues and eigenvectors.

    energies: array of energies over which specteral density is calculated
    current_operator: kwant current operator
    eigen_values, eigen_vector: eigenvalues and eigenvectors from
    sparse diagonilization
    num_moments: number of moments used in the Chebyshev expansion
    bounds: Energy bounds from the kpm spectrum to rescale energies
    in the range [-1, 1].
    """
    a, b = bounds
    rescaled_energy = (energies - b)/a
    eigen_values = (eigen_values - b)/a
    g_e = (np.pi * np.sqrt(1 - rescaled_energy)
                   * np.sqrt(1 + rescaled_energy))
    m = np.arange(num_moments)
    
    current_operator_expectation = np.array([np.sum(current_operator(eigvec)) for eigvec in eigen_vectors.T])
    cns = np.array([nth_moment(n, eigen_values, current_operator_expectation) for n in m])/a
    kernel = ((num_moments - m + 1) *
          np.cos(np.pi * m/(num_moments + 1)) +
          np.sin(np.pi * m/(num_moments + 1)) /
          np.tan(np.pi/(num_moments + 1)))
    kernel = kernel / (num_moments + 1)

    coef_cheb = np.transpose(cns.transpose() * kernel)
    coef_cheb[1:] = 2 * coef_cheb[1:]
    current_densities = np.transpose(np.polynomial.chebyshev.chebval(
        rescaled_energy, coef_cheb) / g_e).real
    return current_densities

def left_cut(site_to, site_from, pos0=0, pos1=1):
    return site_from.pos[0] == pos0 and site_to.pos[0] == pos1

def indices_cutpos(syst, pos0=0, pos1=1):
    cut_indices = []
    for i, site in enumerate(syst.sites):
        if site.pos[0] == pos0 or site.pos[0] == pos1:
            temp = [2 * i, 2 * i + 1]
            cut_indices.append(temp)
    return np.hstack(cut_indices)

def cut_sites(site, pos0=0, pos1=1):
    return site.pos[0] == pos0 or site.pos[0] == pos1

def supercurrent_and_densities(syst, params, k, sigma, num_moments=None, energy_resolution=None,
                       densities=False, return_both=False):
    eigvals, eigvecs = andreev_vectors(syst, params, k=k, sigma=sigma)
    eigvecs_negative = eigvecs[:, eigvals < 0]
    eigvals_negative = eigvals[eigvals < 0]
    if num_moments is None and energy_resolution is None:
        energy_resolution = np.max(abs(eigvals))/5
    kwant_operator = kwant.operator.Current(syst, sz, where=left_cut)
    kwant_operator = kwant_operator.bind(params=params)
    sites_local = indices_cutpos(syst)

    kpm_full_spec = kwant.kpm.SpectralDensity(
        syst, operator=kwant_operator, params=params,
        num_vectors=len(sites_local),
        vector_factory=kwant.kpm.LocalVectors(syst, where=cut_sites),
        num_moments=num_moments,
        energy_resolution=energy_resolution
    )

    energies_full, densities_full = kpm_full_spec()
    bounds = (kpm_full_spec._a, kpm_full_spec._b)
    densities_subgap = current_spectrum(energies_full, kwant_operator, eigvals, eigvecs,
                                        kpm_full_spec.num_moments, bounds)
    densities_continuum = len(sites_local) * np.sum(densities_full, axis=1) - densities_subgap
    if densities:
        current_expectation = np.array([kwant_operator(eigvec) for eigvec in eigvecs.T])
        return energies_full, len(sites_local) * np.sum(densities_full, axis=1),\
                densities_subgap, kpm_full_spec.num_moments, eigvals, current_expectation
    if return_both:
        current_continuum = np.trapz(densities_continuum[energies_full < 0], energies_full[energies_full < 0])
        current_subgap = supercurrent_exact(syst, params, eigvecs_negative, cut=left_cut)
        current_subgap = np.sum(current_subgap)
        current_expectation = np.array([kwant_operator(eigvec) for eigvec in eigvecs.T])
        return eigvals, current_expectation, energies_full, len(sites_local) * np.sum(densities_full, axis=1),\
                densities_subgap, kpm_full_spec.num_moments, current_subgap, current_continuum
    current_continuum = np.trapz(densities_continuum[energies_full < 0], energies_full[energies_full < 0])
    current_subgap = supercurrent_exact(syst, params, eigvecs_negative, cut=left_cut)
    current_subgap = np.sum(current_subgap)
    return current_subgap, current_continuum
