import numpy as np

# Sukhorukov potential
# User defined functions for  and build Kwant system


def make_gate_vertices(device, grid, xmin, ymin):
    
    """
    Find gate vertices based on the input in-plane gate geometry
    
    Parameters:
    -----------
    device      (dict)     in-plane gate geometry
    grid        (nm)       grid_spacing used to discretize contiuum model into tight binding hamiltonian
    xmin        (nm)       left boundary of the rectangular kwant system
    xmin        (nm)       top boundary of the rectangular kwant system
    
    Returns:
    --------
    Vertices of each gate
    """

    Lx = device["Lx"]
    Ly = device["Ly"]
    bx = device["bx"]
    by = device["by"]
    px = device["px"]
    py = device["py"]
    gap = device["gap"]
    nqbit = device["nqbit"]

    # define gate regions by specifying it's vertices
    ngate = nqbit * 2 + 1
    vertex = []
    kink = []

    vertex = [
        [int(-grid / 2 - Lx / 2), 650],
        [int(-grid / 2 - Lx / 2), 700],
        [int(-grid / 2 + Lx / 2), 700],
        [int(-grid / 2 + Lx / 2), 650],
        [xmin, ymin],
        [xmin, by],
        [xmin + bx, by],
        [xmin + bx, ymin],
    ]

    kink.append([xmin, 650 - 20])
    kink.append([xmin, 650])
    kink.append([xmin + bx, 650])
    kink.append([xmin + bx, 650 - 20])

    for i in range(2, ngate + 1):

        if i % 2 == 0:
            vertex.append([vertex[len(vertex) - 1][0] + gap, ymin])
            vertex.append(
                [
                    vertex[len(vertex) - 2][0] + gap,
                    vertex[len(vertex) - 3][1] - (by - py),
                ]
            )
            vertex.append(
                [
                    vertex[len(vertex) - 3][0] + gap + px,
                    vertex[len(vertex) - 4][1] - (by - py),
                ]
            )
            vertex.append([vertex[len(vertex) - 4][0] + gap + px, ymin])

        elif i % 2 == 1:
            vertex.append([vertex[len(vertex) - 1][0] + gap, ymin])
            vertex.append(
                [
                    vertex[len(vertex) - 2][0] + gap,
                    vertex[len(vertex) - 3][1] + (by - py),
                ]
            )
            vertex.append(
                [
                    vertex[len(vertex) - 3][0] + gap + bx,
                    vertex[len(vertex) - 4][1] + (by - py),
                ]
            )
            vertex.append([vertex[len(vertex) - 4][0] + gap + bx, ymin])

            kink.append([vertex[len(vertex) - 4][0], 650 - 20])
            kink.append([vertex[len(vertex) - 3][0], 650])
            kink.append([vertex[len(vertex) - 2][0], 650])
            kink.append([vertex[len(vertex) - 1][0], 650 - 20])

    return vertex, kink


# ------------------------------------------------------------------------------------


def make_2deg_potential(device, grid, x, y, voltage, xmin):

    """ Calculate the gate induced electrostatic potential in 2DEG
    based on https://doi.org/10.1063/1.359446
    
    Parameters:
    -----------
    device     (dict)   device parameters that specifies length and width of 
                        different gates in current device
                        
    grid      (int)     grid_spacing for Kwant discretizer to discretize continuous 
                        hamiltonian
                        
    x, y                (X,Y) positions in the gate dielectric where oxide charges are 
                        randomly distributed [nm]. Made with numpy meshgrid.
       
    voltage   (dict)    gate voltages (volts)
       
    xshift    (int)     Shift along the x-axis to ensure system being symmetric [nm]
    
    Returns:
    -------
    electrostatic potential corresponding to (x,y) positions in the device (volts)
    """

    nqbit = device["nqbit"]
    d = device["thickness"]

    Vb = voltage["Vb"]
    Vpl = voltage["Vpl"]
    Vpr = voltage["Vpr"]
    Vl = voltage["Vl"]

    wall = 1
    ngate = nqbit * 2 + 1

    vertex, kink = make_gate_vertices(device, grid, xmin, 0)

    # Finite rectangular gate
    for i in range(0, ngate + wall):

        def g(u, v):
            R = np.sqrt(u ** 2 + v ** 2 + d ** 2)
            return 1.0 / 2.0 / np.pi * np.arctan2(u * v, d * R)

        phi = (
            g(x - vertex[4 * i][0], y - vertex[4 * i][1])
            + g(x - vertex[4 * i][0], vertex[4 * i + 2][1] - y)
            + g(vertex[4 * i + 2][0] - x, y - vertex[4 * i][1])
            + g(vertex[4 * i + 2][0] - x, vertex[4 * i + 2][1] - y)
        )

        if i == 0:
            potential = phi * Vb

        elif i == 1 or i == ngate + wall - 1:
            potential = potential + phi * Vl

        elif i == 2:
            potential = potential + phi * Vpl

        elif i == 4:
            potential = potential + phi * Vpr

        else:
            potential = potential + phi * Vb

    for i in range(0, nqbit + 1):

        def g(u, v):
            R = np.sqrt(u ** 2 + v ** 2 + d ** 2)
            return 1.0 / 2.0 / np.pi * np.arctan2(u * v, d * R)

        phi = (
            g(x - kink[4 * i][0], y - kink[4 * i][1])
            + g(x - kink[4 * i][0], kink[4 * i + 2][1] - y)
            + g(kink[4 * i + 2][0] - x, y - kink[4 * i][1])
            + g(kink[4 * i + 2][0] - x, kink[4 * i + 2][1] - y)
        )

        potential = potential + phi * Vb

    return potential


# ------------------------------------------------------------------------------------


def back_gate(device, grid, x, y, Vg):

    """
    Calculate the electostatic potential in 2DEG due to back gate. This controls the carrier density accumulated in 
    2DEG. Changing the carrier density due to back gate replicates the change in doping strength in GaAs devices.
    
    
    """

    t = 72
    Lx = device["Lx"]
    Ly = device["Ly"]

    vert = [
        [int(-grid / 2 - Lx / 2), -Ly / 2],
        [int(-grid / 2 - Lx / 2), Ly + Ly / 2],
        [int(-grid / 2 + Lx / 2), Ly + Ly / 2],
        [int(-grid / 2 + Lx / 2), -Ly / 2],
    ]

    def g(u, v):
        R = np.sqrt(u ** 2 + v ** 2 + t ** 2)
        return 1.0 / 2.0 / np.pi * np.arctan2(u * v, t * R)

    phi = (
        g(x - vert[0][0], y - vert[0][1])
        + g(x - vert[0][0], vert[2][1] - y)
        + g(vert[2][0] - x, y - vert[0][1])
        + g(vert[2][0] - x, vert[2][1] - y)
    )

    potential = phi * Vg

    return potential
