import kwant
from kwant.continuum import discretize
import scipy.linalg as la
from scipy import sparse
import scipy.constants as constants


import codes.qubit_electrostatics as gate

# ------------------------------------------------------------------------------------


def kwant_system(device, kwant_par):
    
    
    """
    Given the gate geometry, this function generates boundaries for a rectangular region
    
    Parameters:
    ----------
    device    (dict)      Gate geometry
    kwant_par (dict)      Parameters related to kwant system such as grid_spacing, hopping parameter
    
    
    Returns:
    -------
    (dict) minium and maximum x,y positions
    """

    px = device["px"]
    bx = device["bx"]
    grid_spacing = kwant_par["grid_spacing"]
    nqbit = device["nqbit"]
    gap = device["gap"]
    xshift = 0

    xmin = -(gap * nqbit) - (px * nqbit / 2) - (bx * nqbit / 2) - (bx - 1) / 2 - grid_spacing
    xmax = (gap * nqbit) + (px * nqbit / 2) + (bx * nqbit / 2) + (bx - 1) / 2

    # To facilitate symmetry, calculate the edge of the first barrier gate from left
    i = xmin
    while i >= xmin:
        if i % grid_spacing == 0:
            xshift = int(i)
            break
        else:
            i = i + 1

    return xmin, xmax, xshift


# ------------------------------------------------------------------------------------

# build kwant system
def build_system(model, dev_param, mat_param, kwant_param, system):

    """
    To build the kwant system based on defined geometry and parameters
    
    Input:
    -----
    model: (string) represents hamiltonian of scattering region, 
                          left and right leads
                        
    dev_param:   (dict) device parameters that specifies length and width 
                        of different gates in current device
                      
    mat_param:   (dict) material specific parameters such as effective mass, 
                        dielectric constant
    
    kwant_param: (dict) grid spacing in [nm], hopping integral and no of 
                        eigenstates to calculate in sparse diagonalization
    
    system:      (dict) minimum and maximum positions of the box defining 
                        scattering region in Kwant
                        
                        
    Returns:
    -------
    Finalized Kwant system
    
    """

    grid_spacing = kwant_param["grid_spacing"]
    Lx = dev_param["Lx"]
    xmin = system["xmin"]
    xmax = system["xmax"]
    ymin = system["ymin"]
    ymax = system["ymax"]

    template = discretize(model, ("x", "y"), grid=grid_spacing)

    shape_lead = lambda s: ymin < s.pos[1] < ymax

    def rectangle(site):
        (x, y) = site.pos
        return x > xmin and x < xmax and y > ymin and y < ymax

    sys = kwant.Builder()
    sys.fill(template, rectangle, (0, (ymin + ymax) / 2))

    return sys.finalized()


# ------------------------------------------------------------------------------------

# Load sukhorukov potential into Kwant hamiltonian and calculate charge density
def diagonalize(device, kwant_par, system, voltage, xshift, boundary):

    """
    To find discretized tight binding hamiltonian followed by diagonalization to 
    obtain eigenvalues and eigenstates
    
    Parameters:
    ------------
    device      (dict) device parameters that specifies length and width of different 
                       gates in current device
    
    kwant_par   (dict) grid spacing in [nm], hopping integral and no of eigenstates 
                        to calculate in sparse diagonalization
                        
    system             Discretized tight binding system

    voltage     (dict) voltages applied to different gates 
                       'dvb' and 'dvp' is used while calculating effective model using 
                       hybrid kpm
    xshift             Shift along x-axis to ensure system being symmetric [nm]
    
    boundary    (dict)  minimum and maximum positions of the box defining 
                        scattering region in Kwant

    
    Returns:
    --------
    energies: eigenenergies
    modes : eigenfunctions
    ham: Hamiltonian 
    """

    A0 = kwant_par["A0"]
    nstate = kwant_par["n_state"]
    grid = kwant_par["grid_spacing"]

    Vg = voltage["Vg"]

    xmin = boundary["xmin"]
    xmax = boundary["xmax"]

    def potential(*pos):
        bg = gate.back_gate(device, grid, pos[0], pos[1], Vg)

        pg = gate.make_2deg_potential(
            device, grid, pos[0], pos[1], voltage, xshift
        )  # unit: V
        return bg + pg

    ham = system.hamiltonian_submatrix(params=dict(A0=A0, V=potential), sparse=True)
    energies = []
    modes = []

    energies, modes = sparse.linalg.eigsh(ham, k=nstate, which="SA")

    return energies, modes, ham


# -------------------------------------------------------------------------------------
